// Small Breakout game demo for Arduino UNO with 240x320 Color LCD
// This demo is not playable. The racket always follows ball's position.
// Depends on Adafruit TFTLCD-Library https://github.com/adafruit/TFTLCD-Library
// and Adafruit GFX-Library https://github.com/adafruit/Adafruit-GFX-Library
// Posted 8/2/2015 on my blog: http://blog.boochow.com/article/423049468.html

#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_TFTLCD.h> // Hardware-specific library

#define LCD_CS A3
#define LCD_CD A2
#define LCD_WR A1
#define LCD_RD A0
// optional
#define LCD_RESET A4

// Assign human-readable names to some common 16-bit color values:
#define BLACK   0x0000
#define BLUE    0x001F
#define RED     0xF800
#define GREEN   0x07E0
#define CYAN    0x07FF
#define MAGENTA 0xF81F
#define YELLOW  0xFFE0
#define WHITE   0xFFFF

Adafruit_TFTLCD tft(LCD_CS, LCD_CD, LCD_WR, LCD_RD, LCD_RESET);

#define RACKETSIZE 8
#define RACKETLINE 37
#define SCRNWIDTH 30
#define SCRNHEIGHT 40

#define BLOCKWIDTH 3
#define BLOCKHEIGHT 2
#define BLOCKTOP 3
#define BLOCKBOTTOM 12
#define BLOCKLINES ((BLOCKBOTTOM-BLOCKTOP+1)/BLOCKHEIGHT)
#define NBLOCKS (BLOCKLINES*SCRNWIDTH/BLOCKWIDTH)

boolean exist[NBLOCKS];

void BlocksInit()
{
  int8_t i;
  
  for(i=0; i<NBLOCKS; i++) {
    exist[i] = true;
  }
}

uint8_t BlocksLeft()
{
  int8_t i;
  int8_t leftBlocks;
  
  leftBlocks = 0;
  for (i=0; i<NBLOCKS; i++){
    if (exist[i])
      leftBlocks++;
  }
  return leftBlocks;
}

#define TOSCRN(v) (v<<3)
uint16_t colors[BLOCKLINES]={RED,GREEN,BLUE,YELLOW,MAGENTA};

void BlocksDrawAll(){
  int8_t x,y;
  int8_t w,h;
  int8_t c;
  int8_t i;
  
  w=TOSCRN(BLOCKWIDTH);
  h=TOSCRN(BLOCKHEIGHT);
  c=0;
  i=0;
  for(y=BLOCKTOP; y<=BLOCKBOTTOM; y += BLOCKHEIGHT) {
    for(x=0; x<SCRNWIDTH; x += BLOCKWIDTH) {
      if (exist[i++]) {
        tft.fillRect(TOSCRN(x), TOSCRN(y), w, h, colors[c]);
      }
    }
    c++;
  }
}

void BlocksEraseOne(int8_t block) {
  int8_t x,y;
  int8_t w,h;
  int8_t rows;
  
  w=TOSCRN(BLOCKWIDTH);
  h=TOSCRN(BLOCKHEIGHT);
  rows = SCRNWIDTH/BLOCKWIDTH;
  x = (block % rows) * BLOCKWIDTH;
  y = (block / rows) * BLOCKHEIGHT + BLOCKTOP;
  tft.fillRect(TOSCRN(x), TOSCRN(y), w, h, BLACK);
}

int8_t BlocksFind(uint8_t x, uint8_t y)
{
  if (x > SCRNWIDTH) {
    return -1;
  }
  if ((y < BLOCKTOP)||(y > BLOCKBOTTOM)) {
    return -1;
  }
  uint8_t rows=SCRNWIDTH/BLOCKWIDTH;
  return (y-BLOCKTOP)/BLOCKHEIGHT*rows + x/BLOCKWIDTH;
}

int8_t BlocksHit(uint8_t x, uint8_t y, int8_t *vx, int8_t *vy)
{
  int8_t block;
  
  block = BlocksFind(x + *vx, y);
  if ((block >= 0) && exist[block]) {
    *vx = -*vx;
    return block;
  }
  block = BlocksFind(x, y + *vy);
  if ((block >= 0) && exist[block]) {
    *vy = -*vy;
    return block;
  }
  block = BlocksFind(x + *vx, y + *vy);
  if ((block >= 0) && exist[block]) {
    *vx = -*vx;
    *vy = -*vy;
    return block;
  }
  return -1;
}

typedef struct
{
  uint8_t x,y;
  int8_t vx,vy;
} Ball;

#define BallInit(ball) ball = {.x = SCRNWIDTH-1, .y = BLOCKBOTTOM+BLOCKHEIGHT, .vx = -1, .vy = 1}
#define BallDraw(ball,tft,color) tft.fillRect(TOSCRN(ball.x), TOSCRN(ball.y), TOSCRN(1), TOSCRN(1), color)
#define RacketDraw(racket,tft,color) tft.fillRect(TOSCRN(racket), TOSCRN((RACKETLINE+1)), TOSCRN(RACKETSIZE), TOSCRN(1), color)

#define MAX(a,b)  (((a) > (b)) ? (a) : (b))
#define MIN(a,b)  (((a) < (b)) ? (a) : (b))   

#define GAME_ONGOING 1
#define GAME_RESTART 0

uint8_t gameStatus;

int16_t racket=(SCRNWIDTH-RACKETSIZE)>>1;
Ball ball;

void setup(void) {
  tft.reset();  
  uint16_t identifier = tft.readID();
  tft.begin(identifier);
  tft.fillScreen(BLACK);

  gameStatus = GAME_RESTART;
  BallInit(ball);
}

boolean game_restart()
{
  if ((ball.y > BLOCKBOTTOM) && (ball.vy >0)) {
    BlocksInit();
    BlocksDrawAll();
    return true;
  } else {
    return false;
  }
}

void game_ongoing()
{
  int8_t block;

  do {
    block=BlocksHit(ball.x,ball.y,&ball.vx,&ball.vy);
    if (block >= 0) {
      exist[block] = false;
      BlocksEraseOne(block);
    }
  } while (block > 0); 
}

void loop()
{
  switch (gameStatus) {
    case GAME_RESTART:
      if (game_restart()) {
        gameStatus = GAME_ONGOING;
      }
      break;
    case GAME_ONGOING:
      game_ongoing();
      if (BlocksLeft() == 0) {
        gameStatus=GAME_RESTART;
      }
      break;
  }
  
  BallDraw(ball, tft, BLACK);
  ball.x += ball.vx;
  if (ball.x >= SCRNWIDTH - 1 || ball.x < 1) {
    ball.vx = -ball.vx;
  }
  ball.y += ball.vy;
  if (ball.y == RACKETLINE) {
    if (ball.x >= racket && ball.x < racket+RACKETSIZE) {
      ball.vy = -ball.vy;
    }
  }
  if (ball.y == 0) {
    ball.vy = -ball.vy;
  }
  if (ball.y > SCRNHEIGHT) {
    BallInit(ball);
  }
  BallDraw(ball, tft, YELLOW);

  RacketDraw(racket,tft,BLACK);
  racket=(ball.x)-(RACKETSIZE>>1);
  racket=MAX(0, racket);
  racket=MIN(racket, SCRNWIDTH-RACKETSIZE);
  RacketDraw(racket,tft,WHITE);
  
  delay(16);
}

